/********************************

      Quaffing from Vats

********************************/

#include "Global.h"
#include "Util.h"
#include "Object.h"
#include "Hero.h"
#include "Interface.h"

/* Produces random vats for matter compiler. */
shFeature *
shFeature::newVat ()
{
    shFeature *vat = new shFeature ();
    vat->mType = shFeature::kVat;
    vat->mVat.mHealthy = RNG (-3, 3);
    vat->mVat.mRadioactive = RNG (10) > 0;
    vat->mVat.mAnalyzed = RNG (2);
    return vat;
}

/* Creates random vats for hero to find on map. */
void
shMapLevel::addVat (int x, int y)
{
    shFeature *vat = shFeature::newVat ();
    vat->mX = x;
    vat->mY = y;
    if (vat->mVat.mRadioactive) {
        /* Ordinary radioactive vats on map usually are unhealthy. */
        if (RNG (7))  vat->mVat.mHealthy = 0;
    } else {
        /* Non-radioactive vats likely are good. */
        vat->mVat.mHealthy = RNG (-1, 3);
    }
    vat->mVat.mAnalyzed = 0;
    mFeatures.add (vat);
}

/* Returns true if vat dried up. */
bool
shFeature::quaffFromVat (void)
{
    if (Hero.isInShop ())  Hero.quaffFromOwnedVat (this);

    /* Usually you need to pour some nasty stuff in to achieve this. */
    if (mVat.mHealthy <= -4) {
        I->p ("Arrgghh!!  You vomit.");
        if (Hero.sufferDamage (kAttVatPlague))
            Hero.shCreature::die (kKilled, "a filthy sludge vat");
        mVat.mAnalyzed = 1;
        return false;
    }

    enum Outcome {
        kIrradiate,
        kPurify,
        kGainAbility,
        kGainPower,
        kGoodStuff,
        kRandomStuff,
        kBadStuff,
        kVenom
    };

    /* Set up chances for drinking outcome. */
    shVector<int> chances;
    if (mVat.mRadioactive) {
        /* 10% irradiate, 5% purify, 15% mutant power */
        for (int i = 0; i < 2; ++i)
            chances.add (kIrradiate);
        chances.add (kPurify);
        for (int i = 0; i < 3; ++i)
            chances.add (kGainPower);
    } else {
        /* 25% purify, 5% mutant power */
        for (int i = 0; i < 5; ++i)
            chances.add (kPurify);
        chances.add (kGainPower);
    }
    /* Up to 15% gain ability, 30% good effect. */
    for (int i = mini (3, mVat.mHealthy); i > 0; --i) {
        chances.add (kGainAbility);
        chances.add (kGoodStuff);
        chances.add (kGoodStuff);
    }
    /* Up to 15% venom, 30% bad effect. */
    for (int i = maxi (-3, mVat.mHealthy); i < 0; ++i) {
        chances.add (kVenom);
        chances.add (kBadStuff);
        chances.add (kBadStuff);
    }

    /* Fill the rest with random effects. */
    while (chances.count () < 20)
        chances.add (kRandomStuff);

    int dryup = 3; /* One in three chance to dry up afterwards. */

    /* Choose outcome. */
    Outcome outcome = Outcome (chances.get (RNG (20)));
    if (outcome == kRandomStuff)
        outcome = RNG (2) ? kGoodStuff : kBadStuff;

    /* See what it does. */
    switch (outcome) {
    case kRandomStuff: /* Already handled above. */
        break;
    case kIrradiate:
        Hero.mRad += RNG (25, 100);
        I->p ("Ick!  That had a toxic taste!");
        break;
    case kPurify:
        Hero.mRad -= RNG (1, 200);
        Hero.mRad = maxi (0, Hero.mRad);
        if (!Hero.mRad)
            I->p ("You feel purified!");
        else
            I->p ("You feel less contaminated.");
        break;
    case kGainAbility:
        Hero.gainAbility (false, 1);
        mVat.mAnalyzed = 1;
        dryup = 2;
        break;
    case kGainPower:
        Hero.getMutantPower ();
        dryup = 2;
        break;
    case kGoodStuff:
        switch (RNG (4)) {
        case 0:
            Hero.mHP += NDX (4, 6);
            Hero.mHP = mini (Hero.mHP, Hero.mMaxHP);
            I->p ("You feel better!");
            break;
        case 1:
        {
            int amt = RNG (1, 6);
            Hero.mPsiDrain -= amt;
            Hero.mAbil.mPsi += amt;
            I->p ("You feel invigorated!");
            break;
        }
        case 2:
            I->p ("Mmmm... bouncy bubbly beverage!");
            if (Hero.getStoryFlag ("impregnation")) {
                Hero.abortion (1); /* Here message given slightly differs. */
                I->p ("You feel the alien embryo inside you die.");
            }
            break;
        case 3:
            I->p ("Mmmm... hot fun!");
            if (Hero.needsRestoration ())
                Hero.restoration (1 + mini (3, mVat.mHealthy));
            break;
        }
        break;
    case kBadStuff:
        switch (RNG (3)) {
        case 0:
            I->p ("Oops!  You fall in!  You are covered in slime!");
            Hero.damageEquipment (kAttVatCorrosion, kCorrosive);
            I->p ("You climb out of the vat.");
            break;
        case 1:
        {
            int x = Hero.mX, y = Hero.mY;
            shMonster *monster = new shMonster (kMonVatSlime);
            if (!Level->findNearbyUnoccupiedSquare (&x, &y) and
                !Level->putCreature (monster, x, y))
            {
                I->p ("It's alive!");
            } else {
                I->p ("The sludge gurgles!");
                delete monster;
            }
            break;
        }
        case 2:
            I->p ("You are jolted by an electric shock!");
            if (Hero.sufferDamage (kAttVatShock)) {
                Hero.shCreature::die (kKilled, "an improperly grounded sludge vat");
            }
            break;
        }
        break;
    case kVenom:
        switch (RNG (3)) {
        case 0:
            I->p ("This stuff is poisonous!");
            Hero.abortion ();
            if (Hero.sufferDamage (kAttVatPoison)) {
                Hero.shCreature::die (kKilled, "drinking some unhealthy sludge");
            }
            break;
        case 1:
        case 2:
            I->p ("This stuff is filthy!  (-1 max HP)");
            --Hero.mMaxHP;
            if (Hero.mHP > Hero.mMaxHP)
                --Hero.mHP;
            if (Hero.mMaxHP == 0) {
                Hero.shCreature::die (kKilled, "drinking some unhealthy sludge");
            }
            break;
        }
        mVat.mAnalyzed = 1;
        break;
    }

    return !RNG (dryup);
}

void
shHero::quaffFromVat (shFeature *vat)
{
    assert (vat->mX == Hero.mX and vat->mY == Hero.mY);

    if (vat->quaffFromVat () and Hero.mState == kActing) {
        I->p ("The vat dries up!");
        Level->forgetFeature (vat->mX, vat->mY);
        Level->removeFeature (vat);
    }
}

/* Returns true if the hero has observed the origin of effect. */
bool
shFeature::teleportVat (shFeature *vat, shCreature *perpetrator)
{
    if (!vat or vat->mType != kVat)
        return false;
    bool observed = false;
    int newX, newY;
    int oldX = vat->mX, oldY = vat->mY;
    int prevroom = Level->getRoomID (oldX, oldY);
    Level->findSuitableStairSquare (&newX, &newY);
    if (!Level->getFeature (newX, newY)) {
        if (Hero.canSee (oldX, oldY)) {
            I->p ("The vat flickers and disappears!");
            observed = true;
        } else if (Hero.mX == oldX and Hero.mY == oldY) {
            I->p ("The vat seems to be gone.");
            observed = true;
        }
        Level->addVat (newX, newY);
        shFeature *newvat = Level->getFeature (newX, newY);
        newvat->mVat = vat->mVat; /* Transfer all vat properties. */
        Level->removeFeature (vat);
        if (Hero.canSee (newX, newY)) {
            I->p ("Suddenly a vat materializes out of nowhere and falls down!");
        } else if (Hero.mX == newX and Hero.mY == newY) {
            I->p ("You hear something big flying at you from above ...");
        }
        /* See if the vat falls on someone. */
        shCreature *c = Level->getCreature (newX, newY);
        if (c) {
            if (c->sufferDamage (kAttFallingVat)) {
                if (Hero.canSee (c) or c == &Hero) {
                    I->p ("It crushes %s into a pulp!", AN (c));
                    I->pauseXY (c->mX, c->mY);
                } else if (Hero.canHearThoughts (c)) {
                    I->p ("The mind of %s emanates pure terror for an instant.", AN (c));
                    Hero.checkConcentration (); /* Contact with terrified mind. */
                    I->p ("Then thoughts of %s abruptly vanish.", THE (c));
                    I->pauseXY (c->mX, c->mY);
                }
                c->die (kSlain);
            } else if (Hero.canSee (c) or c == &Hero) {
                I->p ("It lands right on %s!", AN (c));
                I->pauseXY (c->mX, c->mY);
                if (c->canSee (&Hero)) {
                    c->newEnemy (&Hero);
                }
            } else if (Hero.canHearThoughts (c)) {
                I->p ("The mind of %s emanates fear for a moment.", AN (c));
                I->pauseXY (c->mX, c->mY);
                if (c->hasTelepathy () and perpetrator == &Hero) {
                    I->p("Somehow %s blames you for this.", THE (c));
                    c->newEnemy (&Hero);
                }
            }
        }
        /* Did the vat belong to a clerkbot? */
        if (perpetrator == &Hero and Level->isInShop (oldX, oldY)) {
            int newroom = Level->getRoomID (newX, newY);
            if (prevroom == newroom) {
                Hero.movedVat (); /* Harmless furniture rearrange. */
            } else {
                Hero.stolenVat (); /* Vat landed outside shop. */
            }
        }
    } else {
        if (Hero.canSee (oldX, oldY)) {
            I->p ("The vat flickers for a moment.");
            observed = true;
        }
    }
    return observed;
}

/* Returns true if the hero has observed the origin of effect. */
bool
shFeature::teleportAcid (shFeature *acidpit, shCreature *perpetrator)
{
    if (!acidpit or acidpit->mType != kAcidPit)
        return false;
    bool observed = false;
    int oldX = acidpit->mX, oldY = acidpit->mY;
    int newX, newY;
    Level->findOpenSquare (&newX, &newY);
    if (!Level->getFeature (newX, newY)) {
        if (Hero.canSee (oldX, oldY)) {
            I->p ("The acid flickers and disappears!");
            observed = true;
        } else if (Hero.mX == oldX and Hero.mY == oldY and Hero.mZ == -1) {
            I->p ("The acid seems to be gone.");
            observed = true;
        }
        acidpit->mType = kPit;

        if (Hero.canSee (newX, newY)) {
            I->p ("Suddenly an acid waterfall pours out of nowhere!");
        } else if (Hero.mX == newX and Hero.mY == newY) {
            I->p ("You hear rushing water!  Or some other liquid ...");
        }
        /* See if the acid pours down on someone. */
        shCreature *c = Level->getCreature (newX, newY);
        if (c) {
            if (c->sufferDamage (kAttAcidWaterfall)) {
                if (Hero.canSee (c) or c == &Hero) {
                    I->p ("%s %s completely dissolved!", THE (c), c->are ());
                    I->pauseXY (c->mX, c->mY);
                } else if (Hero.canHearThoughts (c)) {
                    I->p ("The mind of %s emanates great pain for an instant.", AN (c));
                    Hero.checkConcentration (); /* Contact with mind in especially painful agony. */
                    I->p ("Then thoughts of %s abruptly vanish.", THE (c));
                    I->pauseXY (c->mX, c->mY);
                }
                c->die (kSlain);
            } else if (Hero.canSee (c) or c == &Hero) {
                I->p ("%s %s washed!", THE (c), c->are ());
                I->pauseXY (c->mX, c->mY);
                if (c->canSee (&Hero)) {
                    c->newEnemy (&Hero);
                }
            } else if (Hero.canHearThoughts (c)) {
                I->p ("The mind of %s emanates pain for a moment.", AN (c));
                I->pauseXY (c->mX, c->mY);
                if (c->hasTelepathy () and perpetrator == &Hero) {
                    I->p("Somehow %s blames you for this.", THE (c));
                    c->newEnemy (&Hero);
                }
            }
        }
        /* Pour down on items. */
        Level->areaEffectObjects (&Attacks[kAttAcidWaterfall],
            NULL, newX, newY, perpetrator);
        /* The acid may land in another pit. */
        shFeature *f = Level->getFeature (newX, newY);
        if (f and f->mType == kPit) {
            if (Hero.canSee (newX, newY))
                I->p ("A pit is filled.");
            f->mType = kAcidPit;
        }
    } else {
        if (Hero.canSee (oldX, oldY)) {
            I->p ("The acid flickers for a moment.");
            observed = true;
        }
    }
    return observed;
}
